/*
    Copyright (C) 2000-2003 Paul Davis

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <vector>
#include <string>
#include <cstdlib>
#include <cctype>

#include <sys/stat.h>
#include <ftw.h>
#include <sndfile.h>
#include <signal.h>
#include <unistd.h>

#include <pbd/forkexec.h>
#include <gtk--.h>
#include <ardour/audio_library.h>
#include <ardour/sndfile_helpers.h>
#include <ardour/utils.h>

#include "ardour_ui.h"
#include "library_ui.h"
#include "prompter.h"

#include "i18n.h"

using namespace std;
using namespace ARDOUR;

static int process_node (const char *file, const struct stat *sb, int flag);
static string length2string (const int, const int);

LibraryTree::LibraryTree (string action, bool import, bool multi, bool persist)
	: importable(import),
	  multiable(multi),
	  persistant (persist),
	main_box (false, 3),
	btn_box_top(true, 4),
	btn_box_bottom (true, 4),
	import_box (false, 4),
	import_btn(action),
	add_btn(_("Add to Library...")),
	remove_btn(_("Remove...")),
	find_btn(_("Find...")),
	folder_btn(_("Add Folder")),
	multichan_check (_("Create multi-channel region"), 0.0),
	files_select(_("Add audio file or directory")),
	info_box (0)
{
	set_title (_("ardour: audio library"));
	set_name ("AudioLibraryWindow");

	add (main_box);
	set_border_width (3);

	main_box.pack_start(hbox, false, false);
	main_box.pack_start(btn_box_top, false, false);
	main_box.pack_start(btn_box_bottom, false, false);
	main_box.pack_start(import_box, false, false);

	hbox.pack_start(scroll);
	scroll.set_usize (200, 300);
	scroll.set_policy (GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	scroll.add_with_viewport(tree);
	tree.set_selection_mode(GTK_SELECTION_SINGLE);

	btn_box_top.pack_start(add_btn);
	btn_box_top.pack_start(folder_btn);
	btn_box_top.pack_start(remove_btn);

	btn_box_bottom.pack_start(find_btn);

	import_box.set_homogeneous (true);
	import_box.pack_start (import_btn);

	if (multi) {
		import_box.pack_start(multichan_check);
	}

	remove_btn.set_sensitive (false);
	import_btn.set_sensitive (false);
	multichan_check.set_active(true);
	multichan_check.set_sensitive (false);

	if (persistant) {
		delete_event.connect (slot (*this, &ArdourDialog::wm_close_event));
	} else {
		delete_event.connect (slot (*this, &ArdourDialog::wm_doi_event));
	}

	add_btn.clicked.connect (slot (*this, &LibraryTree::add_btn_clicked));
	folder_btn.clicked.connect (slot(*this, &LibraryTree::folder_btn_clicked));
	remove_btn.clicked.connect (slot(*this, &LibraryTree::remove_btn_clicked));
	import_btn.clicked.connect (slot(*this, &LibraryTree::import_btn_clicked));
	find_btn.clicked.connect (slot (*this, &LibraryTree::find_btn_clicked));

	files_select.hide_fileop_buttons();
	files_select.set_filename("/");
	files_select.get_ok_button()->clicked.connect (slot ( *this, 
				&LibraryTree::file_ok_clicked));
	files_select.get_cancel_button()->clicked.connect (slot ( *this, 
				&LibraryTree::file_cancel_clicked));

	set_default_size (340, 400);

	Library->added_group.connect (slot (*this, &LibraryTree::added_group));
	Library->removed_group.connect (slot (*this, &LibraryTree::removed_group));
	Library->added_member.connect (slot (*this, &LibraryTree::added_member));
	Library->removed_member.connect (slot 
									 (*this, &LibraryTree::removed_member));
	
	group_selected = "";
	member_selected = "";

	populate ();
}

LibraryTree::~LibraryTree ()
{
}

void
LibraryTree::added_group (string group, string parent)
{
	using namespace Gtk;

	Tree* parent_tree;
	if (parent.length()) {
		parent_tree = uri_mapping[parent]->get_subtree();
	} else {
		parent_tree = &tree;
	}
	
	TreeItem *item = manage(new TreeItem(Library->get_label(group)));
	Tree_Helpers::ItemList items = parent_tree->tree();
	Tree_Helpers::ItemList::iterator i = items.begin();
	
	list<string> groups;
	Library->get_groups(groups, parent);
	list<string>::iterator j = groups.begin();

	while (i != items.end() && j != groups.end()){
		if ((cmp_nocase(Library->get_label(group),Library->get_label(*j)) <= 0) || 
			!((*i)->get_subtree())){
			
			break;
		}
		++i;
		++j;
	}

	parent_tree->tree().insert (i, *item);
	Tree *subtree = manage(new Tree());
	item->set_subtree (*subtree);

	item->select.connect 
			(bind(slot(*this,&LibraryTree::cb_group_select), item, group));
	
	uri_mapping.insert(map<string, TreeItem*>::value_type(group, item));
	uri_parent.insert(map<string,string>::value_type(group, parent));

	subtree->show();
	item->show();
	
	while (gtk_events_pending()){
		gtk_main_iteration();
	}
}

void
LibraryTree::removed_group (string group)
{
	Gtk::TreeItem* group_item = uri_mapping[group];

	Gtk::Tree* parent_tree;
	if (uri_parent[group].length()) {
		parent_tree = uri_mapping[uri_parent[group]]->get_subtree();
	} else {
		parent_tree = &tree;
	}
	
	parent_tree->tree().remove(*group_item);
	uri_mapping.erase(uri_mapping.find(group));
	uri_parent.erase(uri_parent.find(group));
	
	while (gtk_events_pending()){
		gtk_main_iteration();
	}
}

void
LibraryTree::added_member (string member, string parent)
{
	using namespace Gtk;

	Tree* parent_tree;
	if (parent.length()) {
		parent_tree = uri_mapping[parent]->get_subtree();
	} else {
		parent_tree = &tree;
	}
	
	TreeItem *item = manage(new TreeItem(Library->get_label(member)));
	Tree_Helpers::ItemList items = parent_tree->tree();
	Tree_Helpers::ItemList::iterator i = items.begin();
	
	list<string> members;
	Library->get_members(members, parent);
	list<string>::iterator j = members.begin();

	while (i != items.end() && j != members.end()){
		if (cmp_nocase(Library->get_label(member), Library->get_label(*j)) <= 0){
			break;
		}
		++i;
		++j;
	}
	
	parent_tree->tree().insert (i, *item);

	item->select.connect 
			(bind(slot(*this,&LibraryTree::cb_member_select), item, member));
	
	uri_mapping.insert(map<string, TreeItem*>::value_type(member, item));
	uri_parent.insert(map<string,string>::value_type(member, parent));

	item->show();
	
	while (gtk_events_pending()){
		gtk_main_iteration();
	}
}

void
LibraryTree::removed_member (string member)
{
	Gtk::TreeItem* member_item = uri_mapping[member];

	Gtk::Tree* parent_tree;
	if (uri_parent[member].length()) {
		parent_tree = uri_mapping[uri_parent[member]]->get_subtree();
	} else {
		parent_tree = &tree;
	}
	
	parent_tree->tree().remove(*member_item);
	uri_mapping.erase(uri_mapping.find(member));
	uri_parent.erase(uri_parent.find(member));
	
	while (gtk_events_pending()){
		gtk_main_iteration();
	}
}

void
LibraryTree::populate ()
{
	subpopulate (&tree, group_selected);
}

void
LibraryTree::subpopulate (Gtk::Tree* tree, string group)
{
	using namespace Gtk;

	list<string> groups;
	Library->get_groups(groups, group);
	
	list<string>::iterator i; 

	for (i = groups.begin(); i != groups.end(); ++i) {
		TreeItem *item = 
			manage(new TreeItem(Library->get_label(*i)));
		tree->append (*item);
		Tree *subtree = manage(new Tree());
		item->set_subtree (*subtree);
	
		uri_mapping.insert(map<string, Gtk::TreeItem*>::value_type(*i, item));
		uri_parent.insert(map<string,string>::value_type(*i, group));

		item->select.connect 
				(bind(slot(*this,&LibraryTree::cb_group_select), item, *i));

		subpopulate (subtree, *i);
		subtree->show();
		item->show();
	}

	list<string> members;
	Library->get_members(members, group);
	for (i = members.begin(); i != members.end(); ++i) {
		TreeItem *item = manage(new TreeItem(Library->get_label(*i)));
		tree->append (*item);
		item->show();

		uri_mapping.insert(map<string, Gtk::TreeItem*>::value_type(*i, item));
		uri_parent.insert(map<string,string>::value_type(*i, group));

		item->select.connect 
				(bind(slot(*this,&LibraryTree::cb_member_select), item, *i));
	}
}

void
LibraryTree::add_btn_clicked ()
{
	files_select.show_all();
	ARDOUR_UI::instance()->allow_focus (true);
}

// Gah, too many globals
static string parent_uri;
static vector<string>* old_parent;
static vector<string>* old_parent_uri;

static void clone_ftw(void*);
static int ftw_return;
static Gtk::ProgressBar* bar;

void
LibraryTree::file_ok_clicked ()
{
	files_select.hide_all();
	ARDOUR_UI::instance()->allow_focus (false);

	string* file = new string(files_select.get_filename());
	parent_uri = group_selected;

	Gtk::Window* progress_win = new Gtk::Window();
	progress_win->set_title(_("Importing"));
	progress_win->set_policy(false, false, true);
	Gtk::VBox* main_box = manage(new Gtk::VBox());
	progress_win->add(*main_box);
	bar = manage(new Gtk::ProgressBar());
	bar->set_activity_mode(true);
	bar->set_activity_step(15);
	bar->set_activity_blocks(10);
	main_box->pack_start(*bar);
	Gtk::Button* cancel_btn = manage(new Gtk::Button(_("Cancel")));
	main_box->pack_start(*cancel_btn);
	cancel_btn->clicked.connect (slot (*this, &LibraryTree::cancel_import_clicked));
	progress_win->show_all();
	
	clone_ftw((void*)file);
	
	delete progress_win;
}

void
LibraryTree::cancel_import_clicked()
{
	ftw_return = 1;
}

void
clone_ftw(void* ptr)
{
	string* file = (string*) ptr;
	
	old_parent = new vector<string>;
	old_parent_uri = new vector<string>;
	ftw_return = 0;

	if (ftw (file->c_str(), process_node, 100) < 0){
		warning << compose(_("%1 not added to database"), *file) << endmsg;
	}

	Library->save_groups();
	
	delete old_parent;
	delete old_parent_uri;

	delete file;
}

void
LibraryTree::file_cancel_clicked ()
{
	files_select.hide_all();
	ARDOUR_UI::instance()->allow_focus (false);
}

void
LibraryTree::folder_btn_clicked ()
{
	ArdourPrompter prompter (true);
	prompter.set_prompt(_("Folder name:"));

	prompter.done.connect(Gtk::Main::quit.slot());
	prompter.show_all();

	Gtk::Main::run();

	if (prompter.status == Gtkmmext::Prompter::entered) {
		string name;

		prompter.get_result(name);

		if (name.length()) {
			Library->add_group(name, group_selected);
		}
	}
}

void
LibraryTree::cb_group_select (Gtk::TreeItem* item, string uri)
{
	group_selected = uri;
	member_selected = "";
	remove_btn.set_sensitive(true);
	import_btn.set_sensitive(false);
	multichan_check.set_sensitive(false);
	
	if (info_box) {
		delete info_box;
		info_box = 0;
	}
}

void
LibraryTree::cb_member_select (Gtk::TreeItem* item, string uri)
{
	member_selected = uri;
	group_selected = "";
	remove_btn.set_sensitive(true);

	if (importable) {
		import_btn.set_sensitive(true);
	}
	
	if (multiable) {
		multichan_check.set_sensitive(true);
	}
		
	if (info_box) {
		delete info_box;
		info_box = 0;
	}
	
	info_box = new SoundFileBox(member_selected);
	hbox.pack_start(*info_box);
	info_box->show_all();
}

void
LibraryTree::find_btn_clicked ()
{
	SearchSounds* search = new SearchSounds(importable, multiable);

	search->file_chosen.connect(slot (*this, &LibraryTree::file_found));
	search->show_all();
}

void
LibraryTree::file_found (string uri, bool multi)
{
	EMIT_SIGNAL file_chosen (Library->get_member_filename(uri), multi);
}

void
LibraryTree::remove_btn_clicked ()
{
	if (member_selected != ""){
		Library->remove_member(member_selected);
	} else if (group_selected != ""){
		Library->remove_group(group_selected);
	} else {
		error << _("Should not be reached") << endmsg;
	}

	member_selected = "";
	group_selected = "";

	remove_btn.set_sensitive(false);
	import_btn.set_sensitive(false);
}

string
length2string (const int frames, const int sample_rate)
{
	int secs = (int) (frames / (float) sample_rate);
	int hrs =  secs / 3600;
	secs -= (hrs * 3600);
	int mins = secs / 60;
	secs -= (mins * 60);

	int total_secs = (hrs * 3600) + (mins * 60) + secs;
	int frames_remaining = frames - (total_secs * sample_rate);
	float fractional_secs = (float) frames_remaining / sample_rate;

	char duration_str[32];
	sprintf (duration_str, "%02d:%02d:%05.2f", hrs, mins, (float) secs + fractional_secs);

	return duration_str;
}

int
process_node (const char *file, const struct stat *sb, int flag)
{
	bar->set_value(0.0);
	while (gtk_events_pending()){
		gtk_main_iteration();
	}
	bar->set_value(1.0);

	string s_file(file);

	if (s_file.find("/.") != string::npos){
		return ftw_return;
	}

	if (flag == FTW_D) {
		string::size_type size = s_file.find_last_of('/');
		string label = s_file.substr(++size);
		
		while (!old_parent->empty() 
			&& (s_file.find(old_parent->back()) == string::npos)) {

			parent_uri = old_parent_uri->back();			
			old_parent_uri->pop_back();
			old_parent->pop_back();
		}

		string uri = Library->add_group(label, parent_uri);

		old_parent->push_back(s_file);
		old_parent_uri->push_back(parent_uri);
		parent_uri = uri;

		return ftw_return;
	} else if (flag != FTW_F) {
		return ftw_return;
	}

	// We can't realistically check every file - or can we ?
	char* suffix;
	if ((suffix = strrchr (file, '.')) == 0) {
		return ftw_return;
	}

	if (*(suffix+1) == '\0') {
		return ftw_return;
	}

    if (strcasecmp (suffix+1, "wav") != 0 &&
        strcasecmp (suffix+1, "aiff") != 0 &&
        strcasecmp (suffix+1, "aif") != 0 &&
        strcasecmp (suffix+1, "snd") != 0 &&
        strcasecmp (suffix+1, "au") != 0 &&
        strcasecmp (suffix+1, "raw") != 0 &&
        strcasecmp (suffix+1, "sf") != 0 &&
        strcasecmp (suffix+1, "cdr") != 0 &&
        strcasecmp (suffix+1, "smp") != 0 &&
        strcasecmp (suffix+1, "maud") != 0 &&
        strcasecmp (suffix+1, "vwe") != 0 &&
		strcasecmp (suffix+1, "paf") != 0 &&
		strcasecmp (suffix+1, "voc") != 0) {

        return ftw_return;
    }

	/* OK, it stands a good chance of being a sound file that we
	   might be able to handle.
	*/

	SNDFILE *sf;
	SF_INFO info;
	if ((sf = sf_open ((char *) file, SFM_READ, &info)) < 0) {
		error << compose(_("file \"%1\" could not be opened"), file) << endmsg;
		return ftw_return;
	}
	sf_close (sf);

	string uri = Library->add_member(file, parent_uri);

	Library->set_field(uri, "channels", compose("%1", info.channels));
	Library->set_field(uri, "samplerate", compose("%1", info.samplerate));
	Library->set_field(uri, "resolution", 
					   compose("%1", sndfile_data_width(info.format)));
	Library->set_field(uri, "format", compose("%1", info.format));
	
	return ftw_return;
}

void
LibraryTree::import_btn_clicked ()
{
	EMIT_SIGNAL file_chosen (Library->get_member_filename(member_selected),
							 multichan_check.get_active());
}

static const gchar* selector_titles[] = {
	N_("Field"), 
	N_("Value"), 
	0
};

SoundFileBox::SoundFileBox (string uri)
	:
	uri(uri),
	current_pid(0),
	fields(_fields_refiller, this, internationalize (selector_titles), 
		   false, true),
	main_box (false, 3),
	top_box (true, 4),
	bottom_box (true, 4),
	play_btn(_("Play")),
	add_field_btn(_("Add Field...")),
	remove_field_btn(_("Remove Field"))
{
	set_name ("SoundFileBox");

	pack_start (main_box);
	set_border_width (3);

	main_box.pack_start(label, false, false);
	main_box.pack_start(path, false, false);
	main_box.pack_start(length, false, false);
	main_box.pack_start(format, false, false);
	main_box.pack_start(channels, false, false);
	main_box.pack_start(samplerate, false, false);
	main_box.pack_start(fields, true, true);
	main_box.pack_start(top_box, false, false);
	main_box.pack_start(bottom_box, false, false);

	fields.set_usize(200, 200);

	top_box.set_homogeneous(true);
	top_box.pack_start(add_field_btn);
	top_box.pack_start(remove_field_btn);

	remove_field_btn.set_sensitive(false);

	bottom_box.set_homogeneous(true);
//	bottom_box.pack_start(play_btn);

	if ((playcmd = getenv ("SFDB_PLAYER")) == 0){
		play_btn.set_sensitive(false);
	}

	play_btn.clicked.connect (slot (*this, &SoundFileBox::play_btn_clicked));
	add_field_btn.clicked.connect 
			(slot (*this, &SoundFileBox::add_field_clicked));
	remove_field_btn.clicked.connect 
			(slot (*this, &SoundFileBox::remove_field_clicked));

	fields.selection_made.connect (slot 
								   (*this, &SoundFileBox::field_selected));
	fields.choice_made.connect (slot (*this, &SoundFileBox::field_chosen));
	
	Library->fields_changed.connect (slot 
									 (*this, &SoundFileBox::setup_fields));

	setup_labels (uri);
}

SoundFileBox::~SoundFileBox ()
{
	stop_play();
}

void 
SoundFileBox::_fields_refiller (Gtk::CList &list, void* arg)
{
	((SoundFileBox *) arg)->fields_refiller (list);
}

void
SoundFileBox::fields_refiller (Gtk::CList &clist)
{	
	list<string> flist;
	Library->get_fields(flist);
	list<string>::iterator i;
	
	const gchar *rowdata[3];
	guint row = 0;
	for (i=flist.begin(); i != flist.end(); ++i, ++row){
		if (*i != "channels" && *i != "samplerate" && 
			*i != "resolution" && *i != "format") {

			rowdata[0] = (*i).c_str();
			rowdata[1] = Library->get_field(uri, *i).c_str();
			clist.insert_row (row, rowdata);
			++row;
		}
	}

	clist.column(0).set_auto_resize(true);
 	clist.set_sort_column (0);
 	clist.sort ();
}

void
SoundFileBox::setup_labels (string uri)
{
	SNDFILE *sf;
	SF_INFO sf_info;

	string file = Library->get_member_filename(uri);

	if ((sf = sf_open ((char *) file.c_str(), SFM_READ, &sf_info)) < 0) {
		error << compose(_("file \"%1\" could not be opened"), file) << endmsg;
		return;
	}

	label.set_alignment (0.0f, 0.0f);
	label.set_text ("Label: " + Library->get_label(uri));
	path.set_alignment (0.0f, 0.0f);
	path.set_text ("Path: " + file);

	length.set_alignment (0.0f, 0.0f);
	length.set_text (compose("Length: %1", length2string(sf_info.frames, sf_info.samplerate)));

	format.set_alignment (0.0f, 0.0f);
	format.set_text (compose("Format: %1, %2", 
							 sndfile_major_format(sf_info.format), 
							 sndfile_minor_format(sf_info.format)));

	channels.set_alignment (0.0f, 0.0f);
	channels.set_text (compose("Channels: %1", sf_info.channels));
	
	samplerate.set_alignment (0.0f, 0.0f);
	samplerate.set_text (compose("Samplerate: %1", sf_info.samplerate));
}

void
SoundFileBox::play_btn_clicked ()
{
	stop_play ();

	string cmd = compose("%1 %2", playcmd, Library->get_member_filename(uri));
	cerr << cmd << endl;

/*
	extern char **environ;
	if ((current_pid = forkexec_cmd (cmd.c_str(), environ, 0, 0)) < 0) {
		current_pid = 0;
	}
*/
	system(cmd.c_str());
}

void
SoundFileBox::stop_play ()
{
	if (current_pid) {
		kill (current_pid, SIGKILL);
		current_pid = 0;
	}
}

void
SoundFileBox::add_field_clicked ()
{
	ArdourPrompter prompter (true);
	prompter.set_prompt(_("Field name:"));
	
	prompter.show_all();
	prompter.done.connect(Gtk::Main::quit.slot());
	
	Gtk::Main::run();
	
	if (prompter.status == Gtkmmext::Prompter::entered) {
		string name;
		
		prompter.get_result(name);
		
		if (name.length()) {
			Library->add_field(name);
		}
	}
}

void
SoundFileBox::remove_field_clicked ()
{
	Library->remove_field(selected_field);
	selected_field = "";
	remove_field_btn.set_sensitive(false);
}

void
SoundFileBox::setup_fields ()
{
	fields.rescan();
}

void
SoundFileBox::field_chosen (Gtkmmext::Selector *selector,
							Gtkmmext::SelectionResult *res)
{
	if (res) {
		remove_field_btn.set_sensitive(true);
		selected_field = selector->clist().row(res->row)[0].get_text();
	}
}

void
SoundFileBox::field_selected (Gtkmmext::Selector *selector,
							  Gtkmmext::SelectionResult *res)
{	
	if (!res){
		return;
	}

	string field_name(selector->clist().row(res->row)[0].get_text());
	
	ArdourPrompter prompter (true);
	prompter.set_prompt(_("Field value:"));
	prompter.set_initial_text (selector->clist().row(res->row)[1].get_text());
	
	prompter.show_all();
	prompter.done.connect(Gtk::Main::quit.slot());

	Gtk::Main::run();
	
	if (prompter.status == Gtkmmext::Prompter::entered) {
		string data;

		prompter.get_result(data);
		Library->set_field(uri, field_name, data);
	}

	fields.rescan();
}

SearchSounds::SearchSounds (bool import, bool multi)
	: importable(import),
	  multiable(import),
	find_btn (_("Find")),
	and_rbtn (_("AND")),
	or_rbtn (_("OR")),
	fields(_fields_refiller, this, internationalize(selector_titles))
{
	set_title (_("ardour: locate soundfiles"));
	set_name ("AudioSearchSound");

	add(main_box);

	or_rbtn.set_group(and_rbtn.group());
	or_rbtn.set_active(true);
	rbtn_box.set_homogeneous(true);
	rbtn_box.pack_start(and_rbtn);
	rbtn_box.pack_start(or_rbtn);

	bottom_box.set_homogeneous(true);
	bottom_box.pack_start(find_btn);

	fields.set_usize(200, 200);

	main_box.pack_start(fields);
	main_box.pack_start(rbtn_box, false, false);
	main_box.pack_start(bottom_box, false, false);

	delete_event.connect (slot (*this, &ArdourDialog::wm_doi_event));

	find_btn.clicked.connect (slot (*this, &SearchSounds::find_btn_clicked));
	fields.selection_made.connect (slot 
								   (*this, &SearchSounds::field_selected));

	show_all();
}

SearchSounds::~SearchSounds ()
{
}

void 
SearchSounds::_fields_refiller (Gtk::CList &list, void* arg)
{
	((SearchSounds *) arg)->fields_refiller (list);
}

void
SearchSounds::fields_refiller (Gtk::CList &clist)
{	
	list<string> flist;
	Library->get_fields(flist);
	list<string>::iterator i;
	
	const gchar *rowdata[3];
	guint row;
	for (row=0,i=flist.begin(); i != flist.end(); ++i, ++row){
		rowdata[0] = (*i).c_str();
		rowdata[1] = "";
		clist.insert_row (row, rowdata);
	}

	clist.column(0).set_auto_resize(true);
 	clist.set_sort_column (0);
 	clist.sort ();
}

void
SearchSounds::field_selected (Gtkmmext::Selector *selector,
							  Gtkmmext::SelectionResult *res)
{	
	if (!res){
		return;
	}
	
	ArdourPrompter prompter (true);
	prompter.set_prompt(_("Field value:"));
	
	prompter.show_all();
	prompter.done.connect(Gtk::Main::quit.slot());

	Gtk::Main::run();
	
	if (prompter.status == Gtkmmext::Prompter::entered) {
		string data;

		prompter.get_result(data);
		selector->clist().cell(res->row, 1).set_text(data);
	}
}

void
SearchSounds::find_btn_clicked ()
{
	using namespace Gtk::CList_Helpers;
	typedef map<string,string> StringMap;

	StringMap search_info;

	RowList rows = fields.clist().rows();
	RowList::const_iterator i;

	for (i = rows.begin(); i != rows.end(); ++i) {
		Cell cfield((*i)[0]);
		Cell cdata((*i)[1]);
		if (cdata.get_text().length()){
			search_info.insert(
				StringMap::value_type(cfield.get_text(), cdata.get_text()));
		}
	}

	SearchResults* results;
	if (and_rbtn.get_active()){
		results = new SearchResults(search_info, true, importable, multiable);
	} else {
		results = new SearchResults(search_info, false, importable, multiable);
	}

	results->file_chosen.connect (slot (*this, &SearchSounds::file_found));
	results->show_all();
}

void
SearchSounds::file_found (string uri, bool multi)
{
	EMIT_SIGNAL file_chosen (uri, multi);
}

static const gchar* result_titles[] = { 
	N_("Results"), 
	N_("Uris"), 
	0
};

SearchResults::SearchResults (map<string,string> field_values, bool and_search,
							  bool import, bool multi)
	: importable(import),
	  multiable(multi),
	search_info(field_values),
	  search_and (and_search),
	  selection (""),
	  main_box (false, 3),
	  import_box (true, 4),
	  import_btn (_("Import")),
	  multichan_check (_("Create multi-channel region")),
	  results (_results_refiller, this, internationalize (result_titles), false, true)
{
	set_title (_("Ardour: Search Results"));
	set_name ("ArdourSearchResults");
	
	add(main_box);
	set_border_width (3);
	
	main_box.pack_start(hbox);
	hbox.pack_start(results);

	main_box.pack_start(import_box, false, false);
	
	results.set_usize (200, 300);
	
	import_box.set_homogeneous(true);
	import_box.pack_start(import_btn);
	import_box.pack_start(multichan_check);

	import_btn.set_sensitive(false);
	multichan_check.set_active(true);
	multichan_check.set_sensitive(false);
	
	delete_event.connect (slot (*this, &ArdourDialog::wm_doi_event));

	import_btn.clicked.connect (slot (*this, &SearchResults::import_clicked));

	results.choice_made.connect (slot (*this, &SearchResults::result_chosen));

	show_all();
}

SearchResults::~SearchResults ()
{
}

void 
SearchResults::_results_refiller (Gtk::CList &list, void* arg)
{
	((SearchResults *) arg)->results_refiller (list);
}

void 
SearchResults::results_refiller (Gtk::CList &clist)
{
	list<string> results;
	if (search_and) {
		Library->search_members_and (results, search_info);
	} else {
		Library->search_members_or (results, search_info);
	}

	list<string>::iterator i;
	const gchar *rowdata[3];
	guint row;
	for (row=0, i=results.begin(); i != results.end(); ++i, ++row){
		rowdata[0] = (Library->get_label(*i)).c_str();
		// hide the uri in a hidden column
		rowdata[1] = (*i).c_str();
		clist.insert_row (row, rowdata);
	}

	clist.column(1).set_visiblity(false);
	clist.column(0).set_auto_resize(true);
 	clist.set_sort_column (0);
 	clist.sort ();
}

void
SearchResults::import_clicked ()
{
	EMIT_SIGNAL file_chosen(selection, multichan_check.get_active());
}

void 
SearchResults::result_chosen (Gtkmmext::Selector *selector, 
							  Gtkmmext::SelectionResult *res)
{
	if (res) {
		selection = selector->clist().row(res->row)[1].get_text();

		if(importable) {
			import_btn.set_sensitive(true);
		}
		
		if (multiable) {
			multichan_check.set_sensitive(true);
		}
		
		if(info_box){
			delete info_box;
			info_box = 0;
		}
	
		info_box = new SoundFileBox(selection);
		hbox.pack_start(*info_box);
		info_box->show_all();
	}
}
