/*
    Copyright (C) 2003 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: time_axis_view.h,v 1.10 2003/10/20 20:33:24 pbd Exp $
*/

#ifndef __ardour_gtk_time_axis_h__
#define __ardour_gtk_time_axis_h__

#include <gtk--.h>
#include <gtk-canvas.h>

#include <ardour/region.h>

#include "prompter.h"
#include "axis_view.h"

class Editor;
class RegionSelection;

namespace ARDOUR {
	class Session;
	class Region;
	class Session;
	class RouteGroup;
	class Playlist;
}

/**
 * TimeAxisView defines the abstract base class for time-axis views.
 *
 * This class provides the basic LHS controls and display methods. This should be
 * extended to create functional time-axis based views.
 *
 */
class TimeAxisView : public virtual AxisView
{
  public:
	enum TrackHeight { 
                /* canvas units. they need to be odd
		   valued so that there is a precise
		   middle.
		*/
		Largest = 301,
		Large = 201,
		Larger = 101,
		Normal = 51,
		Smaller = 31,
		Small = 21
	};

	/* public data: XXX create accessor/mutators for these ?? */

	Editor& editor;

	guint32 height;  /* in canvas units */
	double  y_position;
	bool    hidden;
   	bool    selected;

	GtkCanvasItem   *canvas_display;
 	Gtk::VBox       *control_parent;

	/* The Standard LHS Controls */
	Gtk::Frame    controls_frame;
	Gtk::EventBox controls_ebox;
	Gtk::HBox     controls_hbox;
	Gtk::VBox     controls_vbox;
	Gtk::HBox     name_hbox;
	Gtk::VBox     name_vbox;
	Gtk::Frame    name_frame;
 	Gtk::Entry    name_entry;

	TimeAxisView(ARDOUR::Session& sess, Editor& ed, Gtk::Widget *canvas);
	virtual ~TimeAxisView ();

	/**
	 * Display this TrackView as the nth component of the parent box, at y.
	 *
	 * @param y 
	 * @param nth
	 * @param parent the parent component
	 * @return the height of this TrackView
	 */
	virtual guint32 show_at (double y, int nth, Gtk::VBox *parent);

	/**
	 * Hides this TrackView
	 */
	virtual void hide ();

	virtual void set_selected (bool);

	/**
	 * Sets the height of this TrackView to one of ths TrackHeghts
	 *
	 * @param h the TrackHeight value to set
	 */
	virtual void set_height (TrackHeight h);

	/**
	 * Steps through the defined TrackHeights for this TrackView.
	 * Sets bigger to true to step up in size, set to fals eot step smaller.
	 *
	 * @param bigger true if stepping should increase in size, false otherwise
	 */
	virtual void step_height (bool bigger);

	virtual ARDOUR::RouteGroup* edit_group() const { return 0; }
	virtual ARDOUR::Playlist* playlist() const { return 0; }

	virtual void set_samples_per_unit (double) = 0;
	virtual void show_selection (TimeSelection&) = 0;
	virtual void hide_selection () = 0;
	virtual void reshow_selection (TimeSelection&) = 0;
	virtual void set_selected_regionviews (RegionSelection&) = 0;
	virtual void show_timestretch (TimeSelection&) = 0;
	virtual void hide_timestretch () = 0;

	virtual ARDOUR::Region* find_next_region (jack_nframes_t pos, ARDOUR::Region::Point, int dir) {
		return 0;
	}

  protected:

	string controls_base_unselected_name;
	string controls_base_selected_name;

	/**
	 * Handle mouse press on our LHS control name entry.
	 *
	 * @param ev the event
	 */
	virtual gint name_entry_button_press (GdkEventButton *ev);

	/**
	 * Handle mouse relaese on our LHS control name entry.
	 * 
	 *@ param ev the event
	 */
	virtual gint name_entry_button_release (GdkEventButton *ev);

	/**
	 * Handle mouse relaese on our LHS control name ebox.
	 * 
	 *@ param ev the event
	 */
	virtual gint controls_ebox_button_release (GdkEventButton *ev);

	/**
	 * Displays the standard LHS control menu at when.
	 *
	 * @param when the popup activation time
	 */
	void popup_display_menu (guint32 when);

	/**
	 * Build the standard LHS control menu.
	 * Subclasses should extend this method to add their own menu options.
	 *
	 */
	virtual void build_display_menu ();

	/**
         * Do anything that needs to be done to dynamically reset
	 * the LHS control menu.
	 */
	virtual gint handle_display_menu_map_event (GdkEventAny *ev) { return FALSE; }

	/**
	 * Build the standard LHS control size menu for the default TrackHeight options.
	 *
	 */
	virtual void build_size_menu();

	/**
	 * Displays the standard LHS controls size menu for the TrackHeight.
	 *
	 * @parem when the popup activation time
	 */
 	void popup_size_menu(guint32 when);

	/**
	 * Handle the size option of out main menu.
	 * 
	 * @param ev the event
	 */
	gint size_click(GdkEventButton *ev);

	/* The standard LHS Track control popup-menus */

	Gtk::Menu *display_menu;
	Gtk::Menu *size_menu;

	Gtk::EventBox expansion_box;
	Gtk::Arrow    expansion_arrow;
	Gtk::Label    name_label;
	gint expansion_button_release_handler (GdkEventButton*);

}; /* class TimeAxisView */

#endif /* __ardour_gtk_time_axis_h__ */

