/*
    Copyright (C) 2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: playlist.cc,v 1.66 2003/10/26 22:05:23 essej Exp $
*/

#include <set>
#include <fstream>

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <limits.h>

#include <sigc++/bind.h>

#include <pbd/failed_constructor.h>
#include <pbd/xml++.h>

#include <ardour/playlist.h>
#include <ardour/session.h>

#include <pbd/stl_delete.h>

#include "i18n.h"

using namespace ARDOUR;
using namespace SigC;

SigC::Signal1<void,Playlist*> Playlist::PlaylistCreated;

struct ShowMeTheList {
    ShowMeTheList (Playlist *pl, const string& n) : playlist (pl), name (n) {}
    ~ShowMeTheList () { 
	    // cerr << ">>>>" << name << endl; playlist->dump(); cerr << "<<<<" << name << endl << endl; 
    };
    Playlist *playlist;
    string name;
};

struct RegionSortByLayer {
    bool operator() (Region *a, Region *b) {
	    return a->layer() < b->layer();
    }
};

Playlist::Playlist (Session& sess, string nom, bool hide)
	: _session (sess)
{
	init (hide);
	_name = nom;
	_orig_diskstream_id = 0;
	
	EMIT_SIGNAL PlaylistCreated (this);

	save_state (_("initial state"));
}

Playlist::Playlist (Session& sess, const XMLNode& node, bool hide)
	: _session (sess)
{
	init (hide);
	_name = "unnamed"; /* reset by set_state */
	_orig_diskstream_id = 0;
	
	if (set_state (node)) {
		error << _("Playlist: cannot create from state.") << endmsg;
		throw failed_constructor();
	}

	EMIT_SIGNAL PlaylistCreated (this);

	save_state (_("initial state"));
}

Playlist::Playlist (const Playlist& other, string namestr, bool hide)
	: _name (namestr), _session (other._session), _orig_diskstream_id(other._orig_diskstream_id)
{
	init (hide);

	other.copy_regions (regions);

	for (list<Region*>::iterator x = regions.begin(); x != regions.end(); ++x) {
		(*x)->set_playlist (this);
	}

	PlaylistCreated (this);

	save_state (_("initial state"));
}

Playlist::Playlist (const Playlist& other, jack_nframes_t start, jack_nframes_t cnt, string str, bool hide)
	: _name (str), _session (other._session), _orig_diskstream_id(other._orig_diskstream_id)
{
	RegionLock rlock1 (this);
	RegionLock rlock2 (&((Playlist&)other));
	
	jack_nframes_t end = start + cnt - 1;

	init (hide);

	for (RegionList::const_iterator i = other.regions.begin(); i != other.regions.end(); i++) {

		Region   *region;
		Region   *new_region;
		jack_nframes_t offset = 0;
		jack_nframes_t position = 0;
		jack_nframes_t len = 0;
		string    new_name;
		Region::OverlapType overlap;

		region = *i;

		overlap = region->coverage (start, end);

		switch (overlap) {
		case Region::OverlapNone:
			continue;

		case Region::OverlapInternal:
			offset = start - region->position();
			position = 0;
			len = cnt;
			break;

		case Region::OverlapStart:
			offset = 0;
			position = region->position() - start;
			len = region->length() - position;
			break;

		case Region::OverlapEnd:
			offset = start - region->position();
			position = 0;
			len = region->length() - offset;
			break;

		case Region::OverlapExternal:
			offset = 0;
			position = region->position() - start;
			len = region->length();
			break;
		}

		_session.region_name (new_name, region->name(), false);

		new_region = new Region (*region, offset, len, new_name);
		new_region->set_layer (region->layer());

		add_region_internal (new_region, position, true);
	}
	
	/* this constructor does NOT notify others (session) */

	/* no need to sort - they are already in order */

	save_state (_("initial state"));
}

void
Playlist::copy_regions (RegionList& newlist) const
{
	RegionLock rlock (const_cast<Playlist *> (this));

	for (RegionList::const_iterator i = regions.begin(); i != regions.end(); ++i) {
		newlist.push_back (new Region (**i));
	}
}

void
Playlist::init (bool hide)
{
	atomic_set (&block_notifications, 0);
	pending_modified = false;
	pending_length = false;
	pending_layering = false;
	pending_state_changed = false;
	_refcnt = 0;
	_hidden = hide;
	_splicing = false;
	_nudging = false;
	in_set_state = false;
	_edit_mode = _session.get_edit_mode();
	_current_state_id = 0;
	in_flush = false;
	subcnt = 0;
	_read_data_count = 0;
	
	Modified.connect (slot (this, &Playlist::mark_session_dirty));
}

Playlist::Playlist (const Playlist& pl)
	: _session (pl._session)
{
	fatal << _("playlist const copy constructor called") << endmsg;
}

Playlist::Playlist (Playlist& pl)
	: _session (pl._session)
{
	fatal << _("playlist non-const copy constructor called") << endmsg;
}

Playlist::~Playlist ()
{
	GoingAway (this);
	
	set<Region *> all_regions;
	
	/* find every region we've ever used, and add it to the set of 
	   all regions.
	*/

	for (RegionList::iterator x = regions.begin(); x != regions.end(); ++x) {
		all_regions.insert (*x);
	}

	for (StateMap::iterator i = states.begin(); i != states.end(); ++i) {
		for (RegionList::iterator r = (*i)->regions.begin(); r != (*i)->regions.end(); ++r) {
			all_regions.insert (*r);
		}
		delete (*i);
	}

	/* delete every region */

	for (set<Region *>::iterator ar = all_regions.begin(); ar != all_regions.end(); ++ar) {
		delete *ar;
	}
}

void
Playlist::set_name (const string& str)
{
	/* in a typical situation, a playlist is being used
	   by one diskstream and also is referenced by the
	   Session. if there are more references than that,
	   then don't change the name.
	*/

	if (_refcnt > 2) {
		return;
	}

	_name = str; 
	EMIT_SIGNAL NameChanged(); 
}

/***********************************************************************
 CHANGE NOTIFICATION HANDLING
 
 Notifications must be delayed till the region_lock is released. This
 is necessary because handlers for the signals may need to acquire
 the lock (e.g. to read from the playlist).
 ***********************************************************************/

void
Playlist::delay_notifications ()
{
	atomic_inc (&block_notifications);
}

void
Playlist::release_notifications ()
{
	if (atomic_dec_and_test(&block_notifications)) {
		flush_notifications ();
	}
}

void
Playlist::notify_modified ()
{
	if (atomic_read (&block_notifications)) {
		pending_modified = true;
	} else {
		pending_modified = false;
		EMIT_SIGNAL Modified();
	}
}

void
Playlist::notify_state_changed ()
{
	if (atomic_read (&block_notifications)) {
		pending_state_changed = true;
	} else {
		pending_state_changed = false;
		EMIT_SIGNAL StateChanged();
	}
}

void
Playlist::notify_region_removed (Region *r)
{
	if (atomic_read (&block_notifications)) {
		pending_removals.insert (pending_removals.end(), r);
	} else {
		EMIT_SIGNAL RegionRemoved (r);
		EMIT_SIGNAL Modified ();
	}
}

void
Playlist::notify_region_added (Region *r)
{
	if (atomic_read(&block_notifications)) {
		pending_adds.insert (pending_adds.end(), r);
	} else {
		EMIT_SIGNAL RegionAdded (r);
		EMIT_SIGNAL Modified ();
	}
}

void
Playlist::notify_length_changed ()
{
	if (atomic_read (&block_notifications)) {
		pending_length = true;
	} else {
		EMIT_SIGNAL LengthChanged();
		EMIT_SIGNAL Modified ();
	}
}

void
Playlist::flush_notifications ()

{
	RegionList::iterator r;
	RegionList::iterator a;
	unsigned int n = 0;

	if (in_flush) {
		return;
	}

	in_flush = true;

	for (a = pending_adds.begin(); a != pending_adds.end(); ++a) {
		EMIT_SIGNAL RegionAdded (*a);
		n++;
	}
	for (r = pending_removals.begin(); r != pending_removals.end(); ++r) {
		EMIT_SIGNAL RegionRemoved (*r);
		n++;
	}

	if (pending_length) {
		pending_length = false;
		EMIT_SIGNAL LengthChanged();
		n++;
	}

	if (pending_layering) {
		pending_layering = false;
		EMIT_SIGNAL LayeringChanged();
		n++;
	}
	
	if (n || pending_modified) {
		pending_modified = false;
		EMIT_SIGNAL Modified ();
	}

	if (pending_state_changed) {
		pending_state_changed = false;
		EMIT_SIGNAL StateChanged ();
	}

	pending_adds.clear ();
	pending_removals.clear ();

	in_flush = false;
}

class RegionSorter 
{
  public:
    RegionSorter (bool splice) : _splice (splice) {}
    
    bool operator() (Region *a, Region *b) {

	    /* Return true means a should come before b, false otherwise */

	    /* in splice mode, we don't care about layers */

	    if (_splice) {
		    return a->position() < b->position();
	    }

	    switch (a->coverage (b->position(), b->position() + b->length())) {

	    case Region::OverlapNone:

		    /* no overlap, so just use position */
		    return a->position() < b->position();

	    case Region::OverlapStart:

		    /* b overlaps the start of a, so b comes first.
		       however, if a is below b, then a must come first
		       to provide correct overlapping read order.
		    */

		    if (a->layer() < b->layer()) {
			    return true;
		    } else {
			    return false;
		    }
		    break;

	    case Region::OverlapEnd:

		    /* b overlaps the end of a, so a would come first.
		       however, if b is below a, then b must come first
		       to provide correct overlapping read order.
		    */

		    if (b->layer() < a->layer()) {
			    return false;
		    } else {
			    return true;
		    }
		    break;

	    case Region::OverlapInternal:

		    /* b is within a. normally, a would come first. 
		       if b is below a, then b must come first
		       to provide correct overlapping read order
		    */

		    if (b->layer() < a->layer()) {
			    return false;
		    } else {
			    return true;
		    }

	    case Region::OverlapExternal:

		    /* b surrounds a. normally, b would come first.
		       however, if a is below it, then a should
		       come first to provide correct overlapping read order
		    */

		    if (a->layer() < b->layer()) {
			    return true;
		    } else {
			    return false;
		    }

	    default:
		    fatal << _("programming error: unknown region coverage type") << endmsg;
		    /*NOTREACHED*/
		    return false;
	    }
    }

  private:
    bool _splice;
};

/*************************************************************
  PLAYLIST OPERATIONS
 *************************************************************/

void
Playlist::add_region (const Region& region, jack_nframes_t position, float times, bool with_save) 
{ 
	RegionLock rlock (this);

	times = fabs (times);

	int itimes = (int) floor (times);
	jack_nframes_t pos = position;

	while (itimes--) {
		Region *copy = new Region (region);
		add_region_internal (copy, pos, true);
		pos += copy->length();
	}
	
	if (floor (times) != times) {
		jack_nframes_t length = (jack_nframes_t) floor (region.length() * (times - floor (times)));
		string name;
		_session.region_name (name, region.name(), false);
		Region *sub = new Region (region, 0, length, name, region.layer());
		add_region_internal (sub, pos, true);
	}

	if (with_save) {
		save_state (_("add region"));
	}
}

void
Playlist::add_region_internal (Region *region, jack_nframes_t position, bool delay_sort)
{
	ShowMeTheList (this, "add region before");
	RegionSorter cmp (_edit_mode == Splice);
	jack_nframes_t old_length = _get_maximum_extent();

	region->set_playlist (this);
	region->set_position (position, this);
	
	if (region->layer() == 0) {
		region->set_layer (regions.size());
	}

	regions.push_back (region);

	if (!delay_sort) {
		regions.sort (cmp);
		check_overlaps (region);
	}

	region->LayerChanged.connect (slot (this, &Playlist::region_layer_changed));
	region->BoundsChanged.connect (bind (slot (this, &Playlist::region_bounds_changed), region));
	region->MuteChanged.connect (slot (this, &Playlist::region_changed));
	region->OpacityChanged.connect (slot (this, &Playlist::region_changed));
	region->EnvelopeChanged.connect (slot (this, &Playlist::region_changed));
	
	ShowMeTheList (this, "add region after");

	notify_region_added (region);

	if (old_length != _get_maximum_extent()) {
		notify_length_changed ();
	}
}

void
Playlist::replace_region (Region& old, Region& newr, jack_nframes_t pos)
{
	ShowMeTheList (this, "remove region before");
	RegionLock rlock (this);
	remove_region_internal (&old);
	add_region_internal (&newr, pos);
	save_state (_("replace region"));
	ShowMeTheList (this, "remove region after");
}

void
Playlist::remove_region (Region *region)
{
	ShowMeTheList (this, "remove region before");
	RegionLock rlock (this);
	remove_region_internal (region);
	save_state (_("remove region"));
	ShowMeTheList (this, "remove region after");
}

int
Playlist::remove_region_internal (Region *region, bool delay_sort)
{
	RegionList::iterator i;
	jack_nframes_t old_length = _get_maximum_extent();

	for (i = regions.begin(); i != regions.end(); ++i) {
		if (*i == region) {

			regions.erase (i);
			if (!delay_sort) {
				RegionSorter cmp (_edit_mode == Splice);
				regions.sort (cmp);
			}
			remove_overlaps (region);
			notify_region_removed (region);
			if (old_length != _get_maximum_extent()) {
				notify_length_changed ();
			}
			return 0;
		}
	}
	return -1;
}

void
Playlist::partition (jack_nframes_t start, jack_nframes_t end, bool just_top_level)
{
	RegionList thawlist;
	
	partition_internal (start, end, false, thawlist);

	for (RegionList::iterator i = thawlist.begin(); i != thawlist.end(); ++i) {
		(*i)->thaw (Region::BoundsNotification, "separation", true, this);
	}

	save_state (_("separate"));
}

void
Playlist::partition_internal (jack_nframes_t start, jack_nframes_t end, bool cutting, RegionList& thawlist)
{
	RegionLock rlock (this);
	Region *region;
	Region *current;
	string new_name;
	RegionList::iterator tmp;
	Region::OverlapType overlap;
	jack_nframes_t pos1, pos2, pos3, pos4;

	/* need to work from a copy, because otherwise the regions we add during the process
	   get operated on as well.
	*/

	RegionList copy = regions;

	for (RegionList::iterator i = copy.begin(); i != copy.end(); i = tmp) {
		
		tmp = i;
		tmp++;

		current = *i;
		
		if (current->first_frame() == start && current->last_frame() == end) {
			if (cutting) {
				remove_region_internal (current);
			}
			continue;
		}
		
		if ((overlap = current->coverage (start, end)) == Region::OverlapNone) {
			continue;
		}
		
		pos1 = current->position();
		pos2 = start;
		pos3 = end;
		pos4 = current->position() + current->length() - 1;

		if (overlap == Region::OverlapInternal) {
			
			/* split: we need 3 new regions, the front, middle and end.
			   cut:   we need 2 regions, the front and end.
			*/
			
			/*
			                 start                 end
			  ---------------*************************------------
			                 P1  P2              P3  P4
			  SPLIT:
			  ---------------*****++++++++++++++++====------------
			  CUT
			  ---------------*****----------------====------------
			  
			*/

			if (!cutting) {
				
				/* "middle" ++++++ */
				
				_session.region_name (new_name, current->name(), false);
				region = new Region (*current, pos2 - pos1, pos3 - pos2, new_name,
						     regions.size(), Region::Flag(current->_flags|Region::Automatic));
				add_region_internal (region, start, true);
			}
			
			/* "end" ====== */
			
			_session.region_name (new_name, current->name(), false);
			region = new Region (*current, pos3 - pos1, pos4 - pos3, new_name, 
					     regions.size(), Region::Flag(current->_flags|Region::Automatic));

			add_region_internal (region, end, true);
			
		        /* "front" ***** */
				
			current->freeze (Region::BoundsNotification);
			thawlist.push_back (current);
			current->trim_end (pos2, this);

		} else if (overlap == Region::OverlapEnd) {
			
			/*
				                              start           end
				    ---------------*************************------------
				                   P1           P2         P4   P3
                                    SPLIT:						   
				    ---------------**************+++++++++++------------
                                    CUT:						   
				    ---------------**************-----------------------

			*/

			if (!cutting) {
				
				/* end +++++ */
				
				_session.region_name (new_name, current->name(), false);
				region = new Region (*current, pos2 - pos1, pos4 - pos2, new_name, (layer_t) regions.size(),
						     Region::Flag(current->_flags|Region::Automatic));
				add_region_internal (region, start, true);
			}

			/* front ****** */

			current->freeze (Region::BoundsNotification);
			thawlist.push_back (current);
			current->trim_end (pos2, this);


		} else if (overlap == Region::OverlapStart) {
			
			/* split: we need 2 regions: the front and the end.
			   cut: just trim current to skip the cut area
			*/
				
			/*
				                        start           end
				    ---------------*************************------------
				       P2          P1 P3                   P4          

				    SPLIT:
				    ---------------****+++++++++++++++++++++------------
				    CUT:
				    -------------------*********************------------
				    
			*/

			if (!cutting) {
				
				/* front **** */
				 _session.region_name (new_name, current->name(), false);
				 region = new Region (*current, pos3 - pos1, pos4 - pos3, current->name(),
						      regions.size(), Region::Flag(current->_flags|Region::Automatic));
				 add_region_internal (region, pos1, true);
			} 
			
			/* end */
			
			current->freeze (Region::BoundsNotification);
			thawlist.push_back (current);
			current->trim_front (pos3, this);

		} else if (overlap == Region::OverlapExternal) {

			/* split: no split required.
			   cut: remove the region.
			*/
				
			/*
				       start                                      end
				    ---------------*************************------------
				       P2          P1 P3                   P4          

				    SPLIT:
				    ---------------*************************------------
				    CUT:
				    ----------------------------------------------------
				    
			*/
			
			remove_region_internal (current);
		}
	}
}

Playlist*
Playlist::cut_copy (Playlist* (Playlist::*pmf)(jack_nframes_t, jack_nframes_t,bool), list<AudioRange>& ranges, bool result_is_hidden)
{
	Playlist* ret;
	Playlist* pl;
	jack_nframes_t start;

	if (ranges.empty()) {
		return 0;
	}

	start = ranges.front().start;

	for (list<AudioRange>::iterator i = ranges.begin(); i != ranges.end(); ++i) {

		pl = (this->*pmf)((*i).start, (*i).length(), result_is_hidden);

		if (i == ranges.begin()) {
			ret = pl;
		} else {
			
			/* paste the next section into the nascent playlist,
			   offset to reflect the start of the first range we
			   chopped.
			*/

			ret->paste (*pl, (*i).start - start, 1.0f);
			delete pl;
		}
	}

	if (ret) {
		/* manually notify session of new playlist here
		   because the playlists were constructed without notifying 
		*/
		PlaylistCreated (ret);
	}
	
	return ret;
}

Playlist*
Playlist::cut (list<AudioRange>& ranges, bool result_is_hidden)
{
	Playlist* (Playlist::*pmf)(jack_nframes_t,jack_nframes_t,bool) = &Playlist::cut;
	return cut_copy (pmf, ranges, result_is_hidden);
}

Playlist*
Playlist::copy (list<AudioRange>& ranges, bool result_is_hidden)
{
	Playlist* (Playlist::*pmf)(jack_nframes_t,jack_nframes_t,bool) = &Playlist::copy;
	return cut_copy (pmf, ranges, result_is_hidden);
}

Playlist *
Playlist::cut (jack_nframes_t start, jack_nframes_t cnt, bool result_is_hidden)
{
	Playlist *the_copy;
	RegionList thawlist;
	char buf[32];
	
	snprintf (buf, sizeof (buf), "%lu", ++subcnt);
	string new_name = _name;
	new_name += '.';
	new_name += buf;

	if ((the_copy = new Playlist (*this, start, cnt, new_name, result_is_hidden)) == 0) {
		return 0;
	}
	
	partition_internal (start, start+cnt-1, true, thawlist);
	possibly_splice ();

	for (RegionList::iterator i = thawlist.begin(); i != thawlist.end(); ++i) {
		(*i)->thaw (Region::BoundsNotification, "playlist cut", true, this);
	}

	save_state (_("cut"));

	return the_copy;
}

Playlist *
Playlist::copy (jack_nframes_t start, jack_nframes_t cnt, bool result_is_hidden)
{
	char buf[32];
	
	snprintf (buf, sizeof (buf), "%lu", ++subcnt);
	string new_name = _name;
	new_name += '.';
	new_name += buf;

	cnt = min (_get_maximum_extent() - start, cnt);
	return new Playlist (*this, start, cnt, new_name, result_is_hidden);
}

int
Playlist::paste (Playlist& other, jack_nframes_t position, float times)
{
	times = fabs (times);
	jack_nframes_t old_length;

	{
		RegionLock rl1 (this);
		RegionLock rl2 (&other);

		old_length = _get_maximum_extent();
	
		int itimes = (int) floor (times);
		jack_nframes_t pos = position;
		jack_nframes_t shift = other._get_maximum_extent();
		layer_t top_layer = regions.size();

		while (itimes--) {
			for (RegionList::iterator i = other.regions.begin(); i != other.regions.end(); ++i) {
				Region *copy_of_region = new Region (**i);

				/* put these new regions on top of all existing ones, but preserve
				   the ordering they had in the original playlist.
				*/
				
				copy_of_region->set_layer (copy_of_region->layer() + top_layer);
				add_region_internal (copy_of_region, copy_of_region->position() + pos);
			}
			pos += shift;
		}

		/* XXX shall we handle fractional cases at some point? */

		if (old_length != _get_maximum_extent()) {
			notify_length_changed ();
		}
	}

	save_state (_("paste"));

	return 0;
}


void
Playlist::duplicate (Region& region, jack_nframes_t position, float times)
{
	times = fabs (times);

	RegionLock rl (this);
	int itimes = (int) floor (times);
	jack_nframes_t pos = position;

	while (itimes--) {
		Region *copy = new Region (region);
		add_region_internal (copy, pos, true);
		pos += region.length();
	}

	if (floor (times) != times) {
		jack_nframes_t length = (jack_nframes_t) floor (region.length() * (times - floor (times)));
		string name;
		_session.region_name (name, region.name(), false);
		Region *sub = new Region (region, 0, length, name);
		add_region_internal (sub, pos, true);
	}

	RegionSorter cmp (_edit_mode == Splice);
	regions.sort (cmp);

	save_state (_("duplicate"));
}

void
Playlist::split_region (Region& region, jack_nframes_t playlist_position)
{
	RegionLock rl (this);

	if (!region.covers (playlist_position)) {
		return;
	}

	if (region.position() == playlist_position ||
	    region.last_frame() == playlist_position) {
		return;
	}

	if (remove_region_internal (&region, true)) {
		return;
	}

	Region *left;
	Region *right;
	jack_nframes_t before;
	jack_nframes_t after;
	string before_name;
	string after_name;

	before = playlist_position - region.position();
	after = region.length() - before;
	
	_session.region_name (before_name, region.name(), false);
	left = new Region (region, 0, before, before_name);

	_session.region_name (after_name, region.name(), false);
	right = new Region (region, before, after, after_name);
	
	add_region_internal (left, region.position(), true);
	add_region_internal (right, region.position() + before);

	save_state (_("split"));
}

void
Playlist::rebuild_overlaps ()
{
	RegionList::iterator i;

	_overlaps.clear ();

	for (i = regions.begin(); i != regions.end(); ++i) {
		check_overlaps (*i);
	}
}

void
Playlist::remove_overlaps (Region *region)
{
	Overlaps::iterator i, tmp;

	for (i = _overlaps.begin(); i != _overlaps.end(); ) {
		tmp = i;
		tmp++;

		if ((*i)->right == region || (*i)->left == region) {
			delete *i;
			_overlaps.erase (i);
		}
		
		i = tmp;
	}
}

void
Playlist::check_overlaps (Region *region)
{
	RegionList::iterator i;
	Region *other;

	for (i = regions.begin(); i != regions.end(); ++i) {

		if ((other = *i) == region) {
			continue;
		}

		Region::OverlapType           overlap_type;
		pair<Overlaps::iterator,bool> insert_result;
		
		overlap_type = region->coverage (other->position(), other->last_frame());

		switch (overlap_type) {
		case Region::OverlapNone:
		case Region::OverlapInternal:
		case Region::OverlapExternal:
			
			/* note: internal/external overlaps (one region completely within the bounds
			   of another) do not count as "overlaps" for our purposes.
			*/
			
			remove_overlaps (region);
			break;
			
		case Region::OverlapStart:
			insert_result = _overlaps.insert (new Overlap (other, region));
			if (insert_result.second) {
				EMIT_SIGNAL NewRegionOverlap (*(insert_result.first));
			}
			break;
			
		case Region::OverlapEnd:
			insert_result = _overlaps.insert (new Overlap (region, other));
			if (insert_result.second) {
				EMIT_SIGNAL NewRegionOverlap (*(insert_result.first));
			}
			break;
		}
	}

	/* check all existing overlaps */
	
	for (Overlaps::iterator x = _overlaps.begin(); x != _overlaps.end(); ++x) {
		if ((*x)->involves (region)) {
			(*x)->update ();
		}
	}
}

void
Playlist::possibly_splice ()
{
	if (_edit_mode == Splice) {
		splice ();
	}
}

void
Playlist::splice ()
{
	{
		RegionLock rl (this);
		
		_splicing = true;
		
		for (RegionList::iterator i = regions.begin(); i != regions.end(); ++i) {
			
			RegionList::iterator next;
			
			next = i;
			++next;
			
			if (next == regions.end()) {
				break;
			}
			
			(*next)->set_position ((*i)->last_frame() + 1, this);
		}
		
		_splicing = false;
	}

	ShowMeTheList (this, "splice");

	notify_length_changed ();
}

void
Playlist::region_bounds_changed (Region::BoundsChangeInfo what_changed, void *src, Region *region)
{
	if (in_set_state || _splicing || _nudging) {
		return;
	}

	if (what_changed & (Region::PositionChange|Region::LengthChange)) {
		sort_regions ();
		possibly_splice ();
		check_overlaps (region);
		save_state (_("region bounds changed"));
		ShowMeTheList (this, "bounds changed");
		notify_length_changed ();
	}
}

void
Playlist::region_layer_changed ()
{
	if (in_set_state) { 
		return;
	}
	sort_regions ();
	save_state (_("layering changed"));
	notify_modified ();
}

void
Playlist::region_changed ()
{
	if (in_set_state) return;
	save_state (_("region state changed"));
	notify_modified ();
}

void
Playlist::sort_regions ()
{
	RegionLock (this);
	RegionSorter cmp (_edit_mode == Splice);
	regions.sort (cmp);
	ShowMeTheList (this, "sort");
}

void
Playlist::clear (bool with_delete, bool with_save)
{
	ShowMeTheList (this, "clear");
	RegionList::iterator i;
	RegionList tmp;

	{ 
		RegionLock (this);
		tmp = regions;
		regions.clear ();
	}
	
	for (i = tmp.begin(); i != tmp.end(); ++i) {
		notify_region_removed (*i);
		if (with_delete) {
			delete *i;
		}
	}

	for (Overlaps::iterator i = _overlaps.begin(); i != _overlaps.end(); ++i) {
		delete *i;
	}
	_overlaps.clear ();

	if (with_save) {
		save_state (_("clear"));
	}
}

/***********************************************************************
 FINDING THINGS
 **********************************************************************/

Playlist::RegionList *
Playlist::regions_at (jack_nframes_t frame)

{
	RegionLock rlock (this);
	return find_regions_at (frame);
}	

Region *
Playlist::top_region_at (jack_nframes_t frame)

{
	RegionLock rlock (this);
	RegionList *rlist = find_regions_at (frame);
	Region *region = 0;

	if (rlist->size()) {
		RegionSortByLayer cmp;
		rlist->sort (cmp);
		region = rlist->back();
	} 

	delete rlist;
	return region;
}	

Playlist::RegionList *
Playlist::find_regions_at (jack_nframes_t frame)
{
	RegionList *rlist = new RegionList;

	for (RegionList::iterator i = regions.begin(); i != regions.end(); ++i) {
		if ((*i)->covers (frame)) {
			rlist->push_back (*i);
		}
	}

	return rlist;
}

Region*
Playlist::find_next_region (jack_nframes_t frame, Region::Point point, int dir)
{
	RegionLock rlock (this);
	Region* ret = 0;
	jack_nframes_t closest = max_frames;

	for (RegionList::iterator i = regions.begin(); i != regions.end(); ++i) {

		jack_nframes_t distance;
		Region* r = (*i);
		jack_nframes_t pos;

		switch (point) {
		case Region::Start:
			pos = r->first_frame ();
			break;
		case Region::End:
			pos = r->last_frame ();
			break;
		case Region::SyncPoint:
			pos = r->first_frame() + r->sync_offset ();
			break;
		}

		switch (dir) {
		case 1: /* forwards */

			if (pos > frame) {
				if ((distance = pos - frame) < closest) {
					closest = distance;
					ret = r;
				}
			}

			break;

		default: /* backwards */

			if (pos < frame) {
				if ((distance = frame - pos) < closest) {
					closest = distance;
					ret = r;
				}
			}
			break;
		}
	}

	return ret;
}

/***********************************************************************/


jack_nframes_t
Playlist::read (Sample *buf, Sample *mixdown_buffer, float *gain_buffer, jack_nframes_t start,
		jack_nframes_t cnt, unsigned chan_n)
{
	jack_nframes_t ret = cnt;
	jack_nframes_t end;

	/* Note: because we do not modify the playlist here, we only
	   need do a tentative on the region lock.  This
	   is useful, since we want to silent-fill the buffer if we
	   can't obtain the lock.

	   No notifications are sent by anything in the call graph for
	   this function, and so we don't need to worry about delaying them
	   explicitly.
	*/

	/* optimizing this memset() away involves a lot of conditionals
	   that may well cause more of a hit due to cache misses 
	   and related stuff than just doing this here.
	   
	   it would be great if someone could measure this
	   at some point.

	   one way or another, parts of the requested area
	   that are not written to be Region::region_at()
	   for all Regions that cover the area need to be
	   zeroed.
	*/

	memset (buf, 0, sizeof (Sample) * cnt);

	TentativeLockMonitor rm (region_lock, __LINE__, __FILE__);

	if (!rm.locked()) {
		return cnt;
	}

	/* the region list is already sorted by position and/or layer,
	   so we will find regions in the correct order. but see the next
	   comment too...
	*/
	
	end =  start + cnt - 1;
	_read_data_count = 0;

	for (RegionList::iterator i = regions.begin(); i != regions.end(); ++i) {
		
		/* it would be nice to not have to iterate over all regions,
		   but the ordering of them by position *and* layer means
		   that we can't tell when we have reached the end of those
		   that we need to read from.
		   
		   in 98% of all cases, the length of the region list is small.
		   the worrying cases might be percussive playlists, with thousands 
		   of very small regions. hopefully, we'll design some scheme for
		   handling them in some way, since that will likely create other
		   scaling issues.
		*/
		
		Region *r = *i;

		if ((chan_n < r->n_channels()) && (r->coverage (start, end) != Region::OverlapNone)) {
			(*i)->read_at (buf, mixdown_buffer, gain_buffer, start, cnt, chan_n);
			_read_data_count += (*i)->read_data_count();
		} 
	}
	
	return ret;
}

void
Playlist::mark_session_dirty ()
{
	if (!in_set_state && !atomic_read (&block_notifications)) {
		_session.set_dirty();
	}
}

int
Playlist::set_state (const XMLNode& node)
{
	in_set_state = true;

	XMLNode *child;
	XMLNodeList nlist;
	XMLNodeConstIterator niter;
	XMLPropertyList plist;
	XMLPropertyConstIterator piter;
	XMLProperty *prop;
	Region *region;
	string region_name;

	clear ();

	if (node.name() != "Playlist") {
		in_set_state = false;
		return -1;
	}

	plist = node.properties();

	for (piter = plist.begin(); piter != plist.end(); ++piter) {

		prop = *piter;
		
		if (prop->name() == "name") {
			_name = prop->value();
		}
		else if (prop->name() == "orig_diskstream_id") {
			sscanf (prop->value().c_str(), "%Lu", &_orig_diskstream_id);
		}
	}

	nlist = node.children();

	for (niter = nlist.begin(); niter != nlist.end(); ++niter) {

		child = *niter;
		
		if (child->name() == "Region") {

			if ((region = _session.XMLRegionFactory (*child, true)) == 0) {
				error << _("Playlist: cannot create Region from state file") << endmsg;
				in_set_state = false;
				return -1;
			}

			add_region (*region, region->position());
		}
	}

	in_set_state = false;
	save_state (_("reset from stored state"));

	return 0;
}

XMLNode&
Playlist::get_state()
{
	return state(true);
}

XMLNode&
Playlist::get_template()
{
	return state(false);
}

XMLNode&
Playlist::state (bool full_state)
{
	XMLNode *node = new XMLNode ("Playlist");
	char buf[64];
	
	node->add_property ("name", _name);

	snprintf (buf, sizeof(buf), "%Lu", _orig_diskstream_id);
	node->add_property ("orig_diskstream_id", buf);

	if (full_state) {
		RegionLock rlock (this, false);

		for (RegionList::iterator i = regions.begin(); i != regions.end(); ++i) {
			node->add_child_nocopy ((*i)->get_state());
		}
	}

	if(_extra_xml){
		node->add_child_copy (*_extra_xml);
	}

	return *node;
}

void
Playlist::dump ()
{
	Region *r;

	cerr << "Playlist \"" << _name << "\" " << endl
	     << regions.size() << " entries in list"
	     << endl;

	for (RegionList::iterator i = regions.begin(); i != regions.end(); ++i) {
		r = *i;
		cerr << "  " << r->name() << ' ' << r->source().name() << " [" 
		     << r->start() << "+" << r->length() 
		     << "] at " 
		     << r->position()
		     << " on layer "
		     << r->layer ()
		     << endl;
	}
}

jack_nframes_t
Playlist::get_maximum_extent () const
{
	RegionLock (const_cast<Playlist *>(this));
	return _get_maximum_extent ();
}

jack_nframes_t
Playlist::_get_maximum_extent () const
{
	RegionList::const_iterator i;
	jack_nframes_t max_extent = 0;
	jack_nframes_t end = 0;

	for (i = regions.begin(); i != regions.end(); ++i) {
		if ((end = (*i)->position() + (*i)->length()) > max_extent) {
			max_extent = end;
		}
	}

	return max_extent;
}

/* HISTORY OPERATIONS */

void
Playlist::store_state (State *state) const
{
	state->regions = regions;
	state->region_states.clear ();

	for (RegionList::const_iterator i = regions.begin(); i != regions.end(); ++i) {
		state->region_states.push_back ((*i)->get_memento());
	}

	ShowMeTheList (const_cast<Playlist*>(this), "after store state");
}

void
Playlist::restore_state (State *state)
{ 
	{ 
		RegionLock rlock (this);
		
		in_set_state = true;

		regions = state->regions;
		
		for (list<UndoAction>::iterator s = state->region_states.begin(); s != state->region_states.end(); ++s) {
			s->call ();
		}
		
		in_set_state = false;
	}

	/* we can't know for sure that we need to do this, but we also didn't want to 
	   respond to region changed signals during the above iteration, so this
	   seems like as good a guess as any ...
	*/

	sort_regions ();
	rebuild_overlaps ();

	notify_length_changed ();
	notify_state_changed (); // XXX this is a bit of an overload, but it'll do

	ShowMeTheList (this, "after restore");
}

UndoAction
Playlist::get_memento () const
{
	void (Playlist::*pmf)(state_id_t) = &Playlist::set_state;
	return SigC::bind (slot (*(const_cast<Playlist *> (this)), pmf), _current_state_id);
}

void
Playlist::save_state (const string& why)
{
	if (!in_set_state) {
		State* state_now = new State (why);

		store_state (state_now);
		states.push_back (state_now);
		_current_state_id = states.size() - 1;
	}
}

void
Playlist::set_state (state_id_t id)
{
	if (id >= states.size()) {
		fatal << _("programming error: illegal state ID passed to Playlist::set_state()") << endmsg;
		/*NOTREACHED*/
		return;
	}

	restore_state (states[id]);
	_current_state_id = id;
	notify_state_changed ();
}

string
Playlist::bump_name (string name)
{
	string::size_type period;
	string newname;

	if ((period = name.find_last_of ('.')) == string::npos) {
		newname = name;
		newname += ".1";
	} else {
		char buf[32];
		int version;
		
		sscanf (name.substr (period+1).c_str(), "%d", &version);
		snprintf (buf, sizeof(buf), "%d", version+1);
		
		newname = name.substr (0, period+1);
		newname += buf;
	}

	return newname;
}

layer_t
Playlist::top_layer() const
{
	RegionLock rlock (const_cast<Playlist *> (this));
	layer_t top = 0;

	for (RegionList::const_iterator i = regions.begin(); i != regions.end(); ++i) {
		top = max (top, (*i)->layer());
	}
	return top;
}

void
Playlist::set_edit_mode (EditMode mode)
{
	_edit_mode = mode;
}

void
Playlist::raise_region (Region& region)
{
	size_t rsz = regions.size();
	layer_t target = region.layer() + 1U;

	if (target >= rsz) {
		return;
	}

	vacate_layer (target, region, 1);
	region.set_layer (target);
}

void
Playlist::lower_region (Region& region)
{
	if (region.layer() == 0) {
		return;
	}

	layer_t target = region.layer() - 1U;

	vacate_layer (target, region, -1);
	region.set_layer (target);
}

void
Playlist::raise_region_to_top (Region& region)
{
	layer_t target = regions.size();

	/* with N regions, acceptable layers = 0..(N-1) */

	if (target > 0) {
		target--;
	}
	
	if (region.layer() == target) {
		return;
	}

	vacate_layer (target, region, 1);
	region.set_layer (target);
}

void
Playlist::lower_region_to_bottom (Region& region)
{
	if (region.layer() == 0) {
		return;
	}

	vacate_layer (0, region, -1);
	region.set_layer (0);
}

int
Playlist::vacate_layer (layer_t target_layer, Region& region, int dir)
{
	RegionList::iterator i;
	typedef pair<Region*,layer_t> LayerInfo;
	list<LayerInfo> layerinfo;

	{
		RegionLock rlock (const_cast<Playlist *> (this));
		
		for (i = regions.begin(); i != regions.end(); ++i) {
			
			if (&region == *i) {
				continue;
			}

			if ((*i)->layer() == target_layer) {
				LayerInfo newpair;
				
				newpair.first = *i;
				newpair.second = region.layer();
				
				layerinfo.push_back (newpair);
			}
		} 
	}

	/* now reset the layers without holding the region lock */

	for (list<LayerInfo>::iterator x = layerinfo.begin(); x != layerinfo.end(); ++x) {
		x->first->set_layer (x->second);
	}

	return 0;
}

void
Playlist::nudge_after (jack_nframes_t start, jack_nframes_t distance, bool forwards)
{
	RegionList::iterator i;
	jack_nframes_t new_pos;
	bool moved = false;

	_nudging = true;

	{
		RegionLock rlock (const_cast<Playlist *> (this));
		
		for (i = regions.begin(); i != regions.end(); ++i) {

			if ((*i)->position() >= start) {

				if (forwards) {

					if ((*i)->last_frame() > max_frames - distance) {
						new_pos = max_frames - (*i)->length();
					} else {
						new_pos = (*i)->position() + distance;
					}
					
				} else {
					
					if ((*i)->position() > distance) {
						new_pos = (*i)->position() - distance;
					} else {
						new_pos = 0;
					}
				}

				(*i)->set_position (new_pos, this);
				moved = true;
			}
		}
	}

	if (moved) {
		_nudging = false;
		save_state (_("nudged"));
		notify_length_changed ();
	}

	ShowMeTheList (this, "after nudge\n");
}


